# -*- coding: utf-8 -*-
"""
Created on Wed Mar 25 16:54:59 2020

@author: amarmore

Computing spectrogram in different feature description.

It is now moved to base_audio package.
"""

from base_audio.signal_to_spectrogram import *